@@ -14,12 +14,7 @@
  *    notice, this list of conditions and the following disclaimer in the 
  *    documentation and/or other materials provided with the distribution. 
  *
- * 3. All advertising materials mentioning features or use of this software 
- *    must display the following acknowledgement: 
- *      This product includes software developed by Kungliga Tekniska 
- *      Hgskolan and its contributors. 
- *
- * 4. Neither the name of the Institute nor the names of its contributors 
+ * 3. Neither the name of the Institute nor the names of its contributors 
  *    may be used to endorse or promote products derived from this software 
  *    without specific prior written permission. 
  *
@@ -38,7 +33,7 @@
 
 #include "krb5_locl.h"
 
-RCSID("$Id: get_in_tkt.c,v 1.87 1999/05/09 08:49:20 assar Exp $");
+RCSID("$Id: get_in_tkt.c,v 1.92 1999/12/02 17:05:09 joda Exp $");
 
 krb5_error_code
 krb5_init_etype (krb5_context context,
@@ -64,7 +59,7 @@ krb5_init_etype (krb5_context context,
 	;
     *len = i;
     *val = malloc(i * sizeof(int));
-    if (*val == NULL) {
+    if (i != 0 && *val == NULL) {
 	ret = ENOMEM;
 	goto cleanup;
     }
@@ -81,7 +76,7 @@ cleanup:
 static krb5_error_code
 decrypt_tkt (krb5_context context,
 	     krb5_keyblock *key,
-	     unsigned usage,
+	     krb5_key_usage usage,
 	     krb5_const_pointer decrypt_arg,
 	     krb5_kdc_rep *dec_rep)
 {
@@ -304,12 +299,12 @@ make_pa_enc_timestamp(krb5_context context, PA_DATA *pa,
     EncryptedData encdata;
     krb5_error_code ret;
     int32_t sec, usec;
-    unsigned usec2;
+    int usec2;
     krb5_crypto crypto;
     
     krb5_us_timeofday (context, &sec, &usec);
     p.patimestamp = sec;
-    usec2 = usec;
+    usec2         = usec;
     p.pausec      = &usec2;
 
     ret = encode_PA_ENC_TS_ENC(buf + sizeof(buf) - 1,
@@ -358,9 +353,9 @@ add_padata(krb5_context context,
 {
     krb5_error_code ret;
     PA_DATA *pa2;
-    krb5_keyblock *key;
     krb5_salt salt2;
     krb5_enctype *ep;
+    int i;
     
     if(salt == NULL) {
 	/* default to standard salt */
@@ -373,25 +368,26 @@ add_padata(krb5_context context,
 	for (ep = enctypes; *ep != ETYPE_NULL; ep++)
 	    netypes++;
     }
-    while (netypes--) {
-	ret = (*key_proc)(context, *enctypes, *salt, keyseed, &key);
-	if (ret != KRB5_KT_NOTFOUND)
-	    break;
-	enctypes++;
+    pa2 = realloc (md->val, (md->len + netypes) * sizeof(*md->val));
+    if (pa2 == NULL)
+	return ENOMEM;
+    md->val = pa2;
+
+    for (i = 0; i < netypes; ++i) {
+	krb5_keyblock *key;
+
+	ret = (*key_proc)(context, enctypes[i], *salt, keyseed, &key);
+	if (ret)
+	    continue;
+	ret = make_pa_enc_timestamp (context, &md->val[md->len],
+				     enctypes[i], key);
+	krb5_free_keyblock (context, key);
+	if (ret)
+	    return ret;
+	++md->len;
     }
     if(salt == &salt2)
 	krb5_free_salt(context, salt2);
-    if (ret)
-	return ret;
-    pa2 = realloc(md->val, (md->len + 1) * sizeof(*md->val));
-    if(pa2 == NULL)
-	return ENOMEM;
-    md->val = pa2;
-    ret = make_pa_enc_timestamp(context, &md->val[md->len], *enctypes, key);
-    krb5_free_keyblock (context, key);
-    if(ret)
-	return ret;
-    md->len++;
     return 0;
 }
 
@@ -561,6 +557,46 @@ fail:
     return ret;
 }
 
+static int
+set_ptypes(krb5_context context,
+	   KRB_ERROR *error, 
+	   krb5_preauthtype **ptypes,
+	   krb5_preauthdata **preauth)
+{
+    static krb5_preauthdata preauth2;
+    static krb5_preauthtype ptypes2[] = { KRB5_PADATA_ENC_TIMESTAMP, KRB5_PADATA_NONE };
+
+    if(error->e_data) {
+	METHOD_DATA md;
+	int i;
+	decode_METHOD_DATA(error->e_data->data, 
+			   error->e_data->length, 
+			   &md, 
+			   NULL);
+	for(i = 0; i < md.len; i++){
+	    switch(md.val[i].padata_type){
+	    case pa_enc_timestamp:
+		*ptypes = ptypes2;
+		break;
+	    case pa_etype_info:
+		*preauth = &preauth2;
+		ALLOC_SEQ(*preauth, 1);
+		(*preauth)->val[0].type = KRB5_PADATA_ENC_TIMESTAMP;
+		krb5_decode_ETYPE_INFO(context,
+				       md.val[i].padata_value.data, 
+				       md.val[i].padata_value.length,
+				       &(*preauth)->val[0].info,
+				       NULL);
+		break;
+	    }
+	}
+	free_METHOD_DATA(&md);
+    } else {
+	*ptypes = ptypes2;
+    }
+    return(1);
+}
+
 krb5_error_code
 krb5_get_in_cred(krb5_context context,
 		 krb5_flags options,
@@ -586,62 +622,85 @@ krb5_get_in_cred(krb5_context context,
     krb5_kdc_flags opts;
     PA_DATA *pa;
     krb5_enctype etype;
+    const krb5_preauthdata *my_preauth = NULL;
     unsigned nonce;
+    int done;
 
     opts.i = options;
 
     krb5_generate_random_block (&nonce, sizeof(nonce));
     nonce &= 0xffffffff;
 
-    ret = init_as_req (context,
-		       opts,
-		       creds,
-		       addrs,
-		       etypes,
-		       ptypes,
-		       preauth,
-		       key_proc,
-		       keyseed,
-		       nonce,
-		       &a);
-    if (ret)
-	return ret;
+    do {
+	done = 1;
+	ret = init_as_req (context,
+			   opts,
+			   creds,
+			   addrs,
+			   etypes,
+			   ptypes,
+			   preauth,
+			   key_proc,
+			   keyseed,
+			   nonce,
+			   &a);
+	if (my_preauth) {
+	    free_ETYPE_INFO(&my_preauth->val[0].info);
+	    free (my_preauth->val);
+	}
+	if (ret)
+	    return ret;
 
-    ret = encode_AS_REQ ((unsigned char*)buf + sizeof(buf) - 1,
-			 sizeof(buf),
-			 &a,
-			 &req.length);
-    free_AS_REQ(&a);
-    if (ret)
-	return ret;
+	ret = encode_AS_REQ ((unsigned char*)buf + sizeof(buf) - 1,
+			     sizeof(buf),
+			     &a,
+			     &req.length);
+	free_AS_REQ(&a);
+	if (ret)
+	    return ret;
 
-    req.data = buf + sizeof(buf) - req.length;
+	req.data = buf + sizeof(buf) - req.length;
 
-    ret = krb5_sendto_kdc (context, &req, &creds->client->realm, &resp);
-    if (ret)
-	return ret;
+	ret = krb5_sendto_kdc (context, &req, &creds->client->realm, &resp);
+	if (ret)
+	    return ret;
 
-    memset (&rep, 0, sizeof(rep));
-    if((ret = decode_AS_REP(resp.data, resp.length, &rep.kdc_rep, &size))) {
-	/* let's try to parse it as a KRB-ERROR */
-	KRB_ERROR error;
-	int ret2;
+	memset (&rep, 0, sizeof(rep));
+	ret = decode_AS_REP(resp.data, resp.length, &rep.kdc_rep, &size);
+	if(ret) {
+	    /* let's try to parse it as a KRB-ERROR */
+	    KRB_ERROR error;
+	    int ret2;
 
-	ret2 = krb5_rd_error(context, &resp, &error);
-	if(ret2 && resp.data && ((char*)resp.data)[0] == 4)
-	    ret = KRB5KRB_AP_ERR_V4_REPLY;
-	krb5_data_free(&resp);
-	if (ret2 == 0) {
-	    ret = error.error_code;
-	    if(ret_as_reply)
-		ret_as_reply->error = error;
-	    else
-		free_KRB_ERROR (&error);
+	    ret2 = krb5_rd_error(context, &resp, &error);
+	    if(ret2 && resp.data && ((char*)resp.data)[0] == 4)
+		ret = KRB5KRB_AP_ERR_V4_REPLY;
+	    krb5_data_free(&resp);
+	    if (ret2 == 0) {
+		ret = error.error_code;
+		/* if no preauth was set and KDC requires it, give it
+                   one more try */
+		if (!ptypes && !preauth
+		    && ret == KRB5KDC_ERR_PREAUTH_REQUIRED
+#if 0
+			|| ret == KRB5KDC_ERR_BADOPTION
+#endif
+		    && set_ptypes(context, &error, &ptypes, &my_preauth)) {
+		    done = 0;
+		    preauth = my_preauth;
+		    free_KRB_ERROR(&error);
+		    continue;
+		}
+		if(ret_as_reply)
+		    ret_as_reply->error = error;
+		else
+		    free_KRB_ERROR (&error);
+		return ret;
+	    }
 	    return ret;
 	}
-	return ret;
-    }
-    krb5_data_free(&resp);
+	krb5_data_free(&resp);
+    } while(!done);
     
     pa = NULL;
     etype = rep.kdc_rep.enc_part.etype;
